from math import inf
from typing import Dict, List, Tuple, Optional

from .calculate_score import (
    calculate_score,
    calculate_complexity_adjusted_score,
)
from .matrices import AlignMatrix
from .performance import timeit


@timeit()
def fill_align_matrix(
    lambda_values: List[float],
    column_count: int,
    edge_start: int,
    chunk_size: int,
    gap_inits: List[float],
    gap_exts: List[float],
    skip_align_score: float,
    subfams: List[str],
    chroms: List[str],
    starts: List[int],
    stops: List[int],
    sub_matrices: List[Dict[str, int]],
    background_freqs: List[Optional[Dict[str, float]]],
) -> AlignMatrix:
    """
    Fills an alignment score matrix by calculating an alignment score (according
    to crossmatch scoring) for every segment of size `chunk_size` for all
    sequences.

    Scores are based on the surrounding `chunk_size` nucleotides in the
    alignment. Ex: column 15 in the matrix holds the aligment score for
    nucleotides at positons 0 - 30, assuming `chunk_size` = 31.

    Starting and trailing cells are treated differently. For example, column 0
    in the matrix holds the alignment score for nucleotides 0 - 15, column 1
    represents nucleotides 0 - 16, etc. Scores are weighted based on number of
    nucleotides that contribute to the score.

    This algorithm ignores all padded indices (".") in the chroms and subfams
    lists.

    Speed up by computing base score for the first segment, moves to next column
    but adding next chars score to base score and subtracting previous chars
    score from base score. Restarts and recomputes new base score when
    necessary.

    Inputs:

    everything needed for CalcScore()
    edge_start: where alignment starts on the target/chrom sequence
    chunk_size: size of nucletide chunks that are scored
    skip_align_score: alignment score to give the skip state (default = 0)
    subfams: alignments to calculate scores from
    chroms: alignments to calculate scores from
    starts: where in the target sequence the competing alignments start

    Outputs:

    align_matrix: Hash implementation of sparse 2D matrix used in pre-DP calculations.
    Key is tuple[int, int] that maps row, col to the value held in that cell of matrix. Rows
    are  subfamilies in the input alignment file, cols are nucleotide positions in the target sequence.

    >>> chros = ["", "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAT...............", "TAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAT..............."]
    >>> subs = ["", "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA...............", "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA--A..............."]
    >>> strts = [0, 2, 0]
    >>> stps = [0, 39, 39]
    >>> sub_mat = [{"AA":1, "AT":-1, "TA":-1, "TT":1, "..":0}] * 3
    >>> back_freqs = [None] * 3
    >>> m = fill_align_matrix([0.0, 0.1, 0.1], 41, 0, 31, [0, -25, -25], [0, -5, -5], 1.0, subs, chros, strts, stps, sub_mat, back_freqs)
    >>> m
    {(1, 3): 3.1, (1, 4): 3.1, (1, 5): 3.1, (1, 6): 3.1, (1, 7): 3.1, (1, 8): 3.1, (1, 9): 3.1, (1, 10): 3.1, (1, 11): 3.1, (1, 12): 3.1, (1, 13): 3.1, (1, 14): 3.1, (1, 15): 3.1, (1, 16): 3.1, (1, 17): 3.1, (1, 18): 3.1, (1, 19): 3.1, (1, 20): 3.1, (1, 21): 3.1, (1, 22): 3.1, (1, 23): 3.1, (1, 24): 3.1, (1, 25): 2.9000000000000004, (1, 26): 2.8933333333333335, (1, 27): 2.8862068965517245, (1, 28): 2.878571428571429, (1, 29): 2.8703703703703702, (1, 30): 2.861538461538462, (1, 31): 2.8520000000000003, (1, 32): 2.841666666666667, (1, 33): 2.8304347826086955, (1, 34): 2.8181818181818183, (1, 35): 2.804761904761905, (1, 36): 2.7900000000000005, (1, 37): 2.7736842105263158, (1, 38): 2.7555555555555555, (1, 39): 2.735294117647059, (1, 40): 2.7125000000000004, (2, 1): 2.7125000000000004, (2, 2): 2.735294117647059, (2, 3): 2.755555555555556, (2, 4): 2.7736842105263158, (2, 5): 2.79, (2, 6): 2.804761904761905, (2, 7): 2.8181818181818183, (2, 8): 2.830434782608696, (2, 9): 2.841666666666667, (2, 10): 2.8520000000000003, (2, 11): 2.861538461538462, (2, 12): 2.8703703703703702, (2, 13): 2.8785714285714286, (2, 14): 2.8862068965517245, (2, 15): 2.8933333333333335, (2, 16): 2.9000000000000004, (2, 17): 3.1, (2, 18): 3.1, (2, 19): 3.1, (2, 20): 3.1, (2, 21): 3.1, (2, 22): 3.1, (2, 23): 0.5, (2, 24): -0.1, (2, 25): -0.30000000000000004, (2, 26): -0.41333333333333333, (2, 27): -0.5344827586206897, (2, 28): -0.6642857142857143, (2, 29): -0.8037037037037037, (2, 30): -0.9538461538461539, (2, 31): -1.116, (2, 32): -1.291666666666667, (2, 33): -1.482608695652174, (2, 34): -1.6909090909090907, (2, 35): -1.9190476190476191, (2, 36): -2.17, (2, 37): -2.447368421052632, (2, 38): -2.7555555555555555, (2, 39): -3.1, (2, 40): -3.4875000000000003, (0, 0): 1.0, (0, 1): 1.0, (0, 2): 1.0, (0, 3): 1.0, (0, 4): 1.0, (0, 5): 1.0, (0, 6): 1.0, (0, 7): 1.0, (0, 8): 1.0, (0, 9): 1.0, (0, 10): 1.0, (0, 11): 1.0, (0, 12): 1.0, (0, 13): 1.0, (0, 14): 1.0, (0, 15): 1.0, (0, 16): 1.0, (0, 17): 1.0, (0, 18): 1.0, (0, 19): 1.0, (0, 20): 1.0, (0, 21): 1.0, (0, 22): 1.0, (0, 23): 1.0, (0, 24): 1.0, (0, 25): 1.0, (0, 26): 1.0, (0, 27): 1.0, (0, 28): 1.0, (0, 29): 1.0, (0, 30): 1.0, (0, 31): 1.0, (0, 32): 1.0, (0, 33): 1.0, (0, 34): 1.0, (0, 35): 1.0, (0, 36): 1.0, (0, 37): 1.0, (0, 38): 1.0, (0, 39): 1.0, (0, 40): 1.0}
    """

    half_chunk: int = int((chunk_size - 1) / 2)
    align_matrix: Dict[Tuple[int, int], float] = {}

    # chunks can't start on gaps and gaps don't count when getting to the chunk_size nucls
    for i in range(1, len(chroms)):
        subfam_seq: str = subfams[i]
        chrom_seq: str = chroms[i]
        sub_matrix = sub_matrices[i]
        lamb = lambda_values[i]
        gap_init = gap_inits[i]
        gap_ext = gap_exts[i]

        char_complexity_adjustments = calculate_complexity_adjusted_score(
            background_freqs[i], subfam_seq, chrom_seq, lamb
        )

        # starts at the first non '.' char, but offsets it in the matrix based on where
        # the alignments start in the seq - ex: if first alignment in the seq starts at 10,
        # will offset by 10

        seq_index: int = 0  # place in subfam_seq and chrom_seq
        col_index = (
            starts[i] - edge_start + half_chunk + 1
        )  # col in align_matrix

        k = half_chunk
        temp_index = seq_index
        temp_count = 0

        while temp_count < chunk_size - k:
            # stop offset before padding starts
            if chrom_seq[temp_index] == ".":
                break
            if chrom_seq[temp_index] != "-":
                temp_count += 1
            temp_index += 1

        offset: int = temp_index - seq_index

        # normalizes for first non trailing cell
        chrom_slice: str = chrom_seq[seq_index : seq_index + offset]
        subfam_slice: str = subfam_seq[seq_index : seq_index + offset]

        # calculates score for first chunk and puts in align_matrix
        align_score: float = calculate_score(
            gap_ext,
            gap_init,
            subfam_slice,
            chrom_slice,
            "",
            "",
            sub_matrix,
            char_complexity_adjustments,
        )

        align_matrix[i, col_index - k] = lamb * (
            align_score * chunk_size / (chunk_size - k)
        )

        # scores for first part, until we get to full sized chunks
        for k in range(half_chunk - 1, -1, -1):
            if (
                chroms[i][seq_index + offset] != "-"
            ):  # if no new gap introduced, move along seq and add next nucl into score
                if subfams[i][seq_index + offset] == "-":
                    if subfams[i][seq_index + offset - 1] == "-":
                        align_score = align_score + gap_ext
                    else:
                        align_score = align_score + gap_init
                else:
                    align_score = (
                        align_score
                        + sub_matrix[
                            subfams[i][seq_index + offset]
                            + chroms[i][seq_index + offset]
                        ]
                        + char_complexity_adjustments[
                            chroms[i][seq_index + offset]
                        ]
                    )
                align_matrix[i, col_index - k] = lamb * (
                    align_score * chunk_size / (chunk_size - k)
                )

                offset += 1

            else:  # if new gap introduced, recalculate offset and call CalcScore again
                temp_index = seq_index
                temp_count = 0

                while temp_count < chunk_size - k:
                    if chrom_seq[temp_index] == ".":
                        break
                    if chrom_seq[temp_index] != "-":
                        temp_count += 1
                    temp_index += 1

                offset = temp_index - seq_index

                chrom_slice = chrom_seq[seq_index : seq_index + offset]
                subfam_slice = subfam_seq[seq_index : seq_index + offset]

                align_score = calculate_score(
                    gap_ext,
                    gap_init,
                    subfam_slice,
                    chrom_slice,
                    subfams[i][seq_index - 1],
                    chroms[i][seq_index - 1],
                    sub_matrix,
                    char_complexity_adjustments,
                )

                align_matrix[i, col_index - k] = lamb * (
                    align_score * chunk_size / (chunk_size - k)
                )

        col_index += 1
        num_nucls: int = chunk_size  # how many nucls contributed to align score

        # move to next chunk by adding next chars score and subtracting prev chars score
        while seq_index + offset < len(chrom_seq):
            temp_index = seq_index
            temp_count = 0

            # stop when you reach the last col to fill in
            if col_index > stops[i] - edge_start + 1:
                break

            # update offset if removing a gap
            if chrom_seq[seq_index] == "-":
                offset -= 1

            # skip over gap and not calc a score for the matrix
            if chrom_seq[seq_index + 1] != "-":
                # if new gap introduced, or gets rid of old gap, recalc offset, rerun CalcScore
                if (
                    chrom_seq[seq_index + offset] == "-"
                    or chrom_seq[seq_index] == "-"
                ):
                    while temp_count < chunk_size:
                        if chrom_seq[temp_index + 1] == ".":
                            break
                        if chrom_seq[temp_index + 1] != "-":
                            temp_count += 1
                        temp_index += 1

                    offset = temp_index - seq_index

                    chrom_slice = chrom_seq[
                        seq_index + 1 : seq_index + offset + 1
                    ]

                    subfam_slice = subfam_seq[
                        seq_index + 1 : seq_index + offset + 1
                    ]
                    align_score = calculate_score(
                        gap_ext,
                        gap_init,
                        subfam_slice,
                        chrom_slice,
                        subfam_seq[seq_index],
                        chrom_seq[seq_index],
                        sub_matrix,
                        char_complexity_adjustments,
                    )

                    temp_count2: int = 0
                    for nuc in chrom_slice:
                        if nuc == ".":
                            break
                        if nuc != "-":
                            temp_count2 += 1
                    num_nucls = temp_count2

                    if num_nucls <= half_chunk:
                        align_score = -inf

                else:
                    # align_score from previous segment - prev chars score + next chars score

                    # subtracting prev chars score
                    if subfam_seq[seq_index] == "-":
                        num_nucls -= 1
                        if subfam_seq[seq_index - 1] == "-":
                            align_score = align_score - gap_ext
                        else:
                            align_score = align_score - gap_init
                    else:
                        align_score = (
                            align_score
                            - sub_matrix[
                                subfam_seq[seq_index] + chrom_seq[seq_index]
                            ]
                            - char_complexity_adjustments[chrom_seq[seq_index]]
                        )
                        num_nucls -= 1

                    # adding next chars score
                    if subfam_seq[seq_index + offset - half_chunk] == ".":
                        break
                    elif subfam_seq[seq_index + offset] == "-":
                        num_nucls += 1
                        if subfam_seq[seq_index + offset - 1] == "-":
                            align_score = align_score + gap_ext
                        else:
                            align_score = align_score + gap_init
                    elif subfam_seq[seq_index + offset] == ".":
                        align_score = align_score
                    else:
                        align_score = (
                            align_score
                            + sub_matrix[
                                subfam_seq[seq_index + offset]
                                + chrom_seq[seq_index + offset]
                            ]
                            + char_complexity_adjustments[
                                chrom_seq[seq_index + offset]
                            ]
                        )
                        num_nucls += 1

                align_matrix[i, col_index] = lamb * (
                    align_score / num_nucls * chunk_size
                )

                if align_score == -inf:
                    del align_matrix[i, col_index]
                    break
                col_index += 1

            seq_index += 1

    # assigns skip states an alignment score
    # do not lambda adjust skip state score
    for j in range(column_count):
        align_matrix[0, j] = skip_align_score

    return align_matrix



from logging import Logger
from os import environ
from sys import stderr
from time import time
from typing import Optional, TextIO

ENV_VAR_NAME = "POLYA_PERFORMANCE"

_logger: Optional[Logger] = None
_output_file: Optional[TextIO] = stderr


def set_timeit_file(file: TextIO):
    global _output_file
    _output_file = file


def reset_timeit_file():
    global _output_file
    _output_file = stderr


def set_timeit_logger(logger: Logger):
    global _logger
    _logger = logger


def reset_timeit_logger():
    global _logger
    _logger = None


def timeit(name: Optional[str] = None):
    """
    A benchmarking helper intended for us as a decorator. To use, decorate a
    function with `@timeit()`, then set the `POLYA_PERFORMANCE` environment
    variable to any value when the program is run.

    Output will be written to stderr unless a `file` is provided by calling
    `set_timeit_file`, in which case that will be used. If a logger is passed to
    `set_timeit_logger` then that will be used for output preferentially.

    >>> import os, io
    >>> stream = io.StringIO()
    >>> os.environ[ENV_VAR_NAME] = 'true'
    >>> set_timeit_file(stream)
    >>> perf_sum = timeit()(sum)
    >>> perf_sum([1, 2])
    3
    >>> stream.getvalue()[:31]
    '[PERFORMANCE] time spent in sum'
    """

    def decorator(method):
        if ENV_VAR_NAME not in environ:
            return method

        def timed(*args, **kw):
            ts = time()
            result = method(*args, **kw)
            te = time()

            if name is None:
                method_name = method.__name__
            else:
                method_name = name

            msg = f"time spent in {method_name} {float(te - ts):.4f}s"

            if _logger:
                _logger.info(msg)
            elif _output_file:
                _output_file.write(f"[PERFORMANCE] {msg}\n")

            return result

        return timed

    return decorator



from typing import Dict, List, NamedTuple, Tuple

RowColumn = Tuple[int, int]

AlignMatrix = Dict[RowColumn, float]

ConfidenceMatrix = Dict[RowColumn, float]

ConsensusMatrix = Dict[RowColumn, int]

NodeConfidenceMatrix = Dict[Tuple[str, int], float]

StrandMatrix = Dict[RowColumn, str]

SupportMatrix = Dict[RowColumn, float]

SubfamCol = Tuple[str, int]

RowConsensus = Tuple[int, int]

SubfamAlignmentsMatrix = Dict[SubfamCol, RowConsensus]


class ConsensusMatrixContainer(NamedTuple):
    """
    A container for a consensus matrix along with its context.
    """

    active_rows: Dict[int, List[int]]
    """
    Mapping from column index to a list of row
    indices in that column that are "active" (have data in them).
    This lets us avoid looping through unnecessary rows.
    """

    matrix: ConsensusMatrix
    """    
    Hash implementation of a sparse 2D matrix used along with DP
    matrices. The key is a 2-tuple of row and column indices
    (respectively) and the values are the alignment positions in the
    consensus subfamily sequence.
    """


class CollapsedMatrices(NamedTuple):
    row_num_update: int
    """
    all matrices used in DP are "collapsed" meaning if there are duplicate rows with the same
    subfamily they are collapsed into the same row. We choose which original row to use in the
    collapsed version by doing a mini DP to find the most probable.

    updates number of rows in matrices
    """

    consensus_matrix: ConsensusMatrix
    # consensus_matrix_collapse: collapsed version

    strand_matrix: StrandMatrix
    # strand_matrix_collapse: Hash implementation of sparse 2D matrix used along with DP matrices.
    # Tuple[str, int] as key. String is the subfamily name of the row, int is the column in matrix.
    # This is a collapsed matrix with no redundant subfamilies as rows. Each cell in matrix is
    # the strand of the consensus sequence that aligned at the corresponding column position of
    # the sequence.

    support_matrix: SupportMatrix
    # support_matrix_collapse: Collapsed version of Support matrix. There may be duplicate rows of the
    # same subfamily, collapsing the matrices puts the duplicates into the same row. Hash
    # implementation with tuple[str, int] as key. String is the subfamily name of the row, int
    # is the column.

    subfamilies: List[str]
    # subfams_collapse: Collapsed version of the array Subfams, any duplicate subfams are consolidated
    # into one.

    active_rows: Dict[int, List[int]]
    # active_cells_collapse: Not all rows in each column hold values. Dictionary that holds column
    # number as the key, and an array of which rows hold values for that column.

    subfamily_indices: Dict[str, int]
    #     subfams_collapse_temp: maps subfam names to it's new row number

    subfam_alignments_matrix: SubfamAlignmentsMatrix
    # subfam_alignment_matrix: Hash implementation of sparse 2D matrix.
    # Tuple[str, int] as key. String is the subfamily name of the row, int is the sequence position.
    # This is a collapsed matrix with no redundant subfamilies as rows. Each cell in the matrix is
    # the original subfam/chrom alignment row and the consensus sequence pos that aligned at the
    # corresponding column position of the sequence.



from typing import Dict, List, Tuple, Set
from math import inf, log

from .matrices import (
    CollapsedMatrices,
    ConsensusMatrix,
    SupportMatrix,
    SubfamAlignmentsMatrix,
)
from .performance import timeit


@timeit()
def dp_for_collapse(
    dp_rows: List[int],
    active: Dict[int, List[int]],
    support_matrix: SupportMatrix,
    non_empty: List[int],
) -> List[int]:
    """
    When there is more than one row with the same subfam label, do a mini DP trace
    to choose which row to use in collapsed matrices

    input:
    dp_rows: rows in original matrices to be collapsed
    support_matrix: uncollapsed support matrix, use these score for dp
    columns: list of non empty cols in matrices

    output:
    path: list with labels for which row to collapse to for each column
    non_empty: list of non empty cols for this particular subfam (in the mini dp matrix)

    >>> dp_r = [0, 2]
    >>> act = {0:[0,2], 2:[0,2], 3:[0,2]}
    >>> non_cols = [0, 2, 3]
    >>> sup_mat = {(0, 0): 0.5, (0, 2): 0.5, (0, 3): .1, (2, 0): 0.1, (2, 2): 0.1, (2, 3): 0.9}
    >>> p = dp_for_collapse(dp_r, act, sup_mat, non_cols)
    >>> p
    [2, 2, 2]
    """
    change: float = log(0.000000001)
    stay: float = log(1 - change)

    path: List[int] = []

    origin: Dict[Tuple[int, int], int] = {}
    dp: Dict[Tuple[int, int], float] = {}

    map_rows: Dict[int, int] = {}
    for k in range(len(dp_rows)):
        # maps active rows from matrix position to position in dp_rows
        map_rows[dp_rows[k]] = k

    # first col of prob_matrix is 0s
    for i in active[non_empty[0]]:
        dp[i, non_empty[0]] = 0.0

    # do dp and fill origin matrix
    # only touches active cells
    for i in range(1, len(non_empty)):
        curr_col = non_empty[i]
        prev_col = non_empty[i - 1]

        for row0 in active[curr_col]:
            row = map_rows[row0]

            max_value: float = -inf
            max_index: int = -1
            support_log: float = log(support_matrix[row0, curr_col])

            for row1 in active[prev_col]:
                prev_row = map_rows[row1]

                score: float = support_log + dp[row1, prev_col]

                if row == prev_row:
                    score += stay
                else:
                    score += change

                if score > max_value:
                    max_value = score
                    max_index = row1

            origin[row0, curr_col] = max_index
            dp[row0, curr_col] = max_value

    # get path from origin matrix
    # which row to start backtrace
    max_value = -inf
    max_row_index = 0
    for i in active[non_empty[-1]]:
        if max_value < dp[i, non_empty[-1]]:
            max_value = dp[i, non_empty[-1]]
            max_row_index = i

    prev_row_index: int = origin[max_row_index, non_empty[-1]]
    path.append(max_row_index)

    # already added the last col, but this adds the one before cols so still start at last col
    for columns_index in range(len(non_empty) - 1, 1, -1):
        prev_column: int = non_empty[columns_index - 1]
        path.append(prev_row_index)
        prev_row_index = origin[prev_row_index, prev_column]

    path.append(prev_row_index)

    path.reverse()

    return path


@timeit()
def collapse_matrices(
    row_count: int,
    start_all: int,
    columns: List[int],
    subfams: List[str],
    strands: List[str],
    starts: List[int],
    stops: List[int],
    active_cells: Dict[int, List[int]],
    support_matrix: SupportMatrix,
    consensus_matrix: ConsensusMatrix,
) -> CollapsedMatrices:
    """
    In all preceding matrices, collapse and combine rows that are the same subfam.

    input:
    row_num: number of rows in matrices
    columns: list of non empty cols in matrices
    subfams: subfamily names for rows in matrices - each row is a different alignment
    strands: which strand each alignment is on
    active_cells: maps column number with rows that are active in that column
    support_matrix: uncollapsed support matrix - rows are number indices
    consensus_matrix: uncollapsed consensus matrix - rows are number indices

    output:
    CollapsedMatrices container

    >>> non_cols = [1, 2, 3]
    >>> active = {1: [0, 1, 2], 2: [0, 1, 2], 3: [0, 1, 2]}
    >>> subs = ["s1", "s2", "s3"]
    >>> strandss = ["+", "-", "-"]
    >>> sup_mat = {(0, 1): 0.5, (0, 2): 0.5, (0, 3): .1, (1, 1): 0.2, (1, 2): 0.2, (1, 3): .2, (2, 1): 0.1, (2, 2): 0.1, (2, 3): 0.9}
    >>> con_mat = {(0, 1): 0, (0, 2): 1, (0, 3): 2, (1, 1): 0, (1, 2): 1, (1, 3): 2, (2, 1): 0, (2, 2): 3, (2, 3): 10}
    >>> (r, con_mat_col, strand_mat_col, sup_mat_col, sub_col, active_col, sub_col_ind, sub_aligns) = collapse_matrices(3, 1, non_cols, subs, strandss, [0,0,0], [2,2,2], active, sup_mat, con_mat)
    >>> r
    3
    >>> con_mat_col
    {(0, 1): 0, (1, 1): 0, (2, 1): 0, (0, 2): 1, (1, 2): 1, (2, 2): 3, (0, 3): 2, (1, 3): 2, (2, 3): 10}
    >>> strand_mat_col
    {(0, 1): '+', (1, 1): '-', (2, 1): '-', (0, 2): '+', (1, 2): '-', (2, 2): '-', (0, 3): '+', (1, 3): '-', (2, 3): '-'}
    >>> sup_mat_col
    {(0, 1): 0.5, (1, 1): 0.2, (2, 1): 0.1, (0, 2): 0.5, (1, 2): 0.2, (2, 2): 0.1, (0, 3): 0.1, (1, 3): 0.2, (2, 3): 0.9}
    >>> sub_col
    ['s1', 's2', 's3']
    >>> active_col
    {1: [0, 1, 2], 2: [0, 1, 2], 3: [0, 1, 2]}
    >>> sub_col_ind
    {'s1': 0, 's2': 1, 's3': 2}
    >>> sub_aligns
    {('s1', 1): (0, 0), ('s2', 1): (1, 0), ('s3', 1): (2, 0), ('s1', 2): (0, 1), ('s2', 2): (1, 1), ('s3', 2): (2, 3), ('s1', 3): (0, 2), ('s2', 3): (1, 2), ('s3', 3): (2, 10)}
    """
    # fixme - write another test for when there is overlap and mini dp is done

    consensus_matrix_collapse: Dict[Tuple[int, int], int] = {}
    strand_matrix_collapse: Dict[Tuple[int, int], str] = {}
    support_matrix_collapse: Dict[Tuple[int, int], float] = {}
    subfams_collapse: List[str] = []
    subfams_collapse_temp: Dict[
        str, int
    ] = {}  # maps subfam name to new row index
    subfams_count: Dict[str, List[int]] = {}  # original row in matrix
    active_cells_collapsed_set: Dict[int, Set[int]] = {}
    subfams_dp = set()

    subfam_alignments_collapse: SubfamAlignmentsMatrix = {}

    # assigns row num to subfams strings
    subfam_row_index: int = 0
    for i in range(row_count):
        subfam = subfams[i]
        if subfam in subfams_count:
            if subfam != "Tandem Repeat":
                subfams_count[subfam].append(i)
                # Only do the DP with non-TRs since TRs won't overlap
                subfams_dp.add(subfam)
        else:
            subfams_collapse.append(subfam)
            subfams_count[subfam] = [i]
            subfams_collapse_temp[subfam] = subfam_row_index
            subfam_row_index += 1

    for col_index in columns:
        # use set so don't add duplicates of subfams with mulitple alignments
        active_cells_collapsed_set[col_index] = set()

        for row_index in active_cells[col_index]:
            subfam = subfams[row_index]

            # if there is only one alignment for a subfam, just copy what was in the uncollapsed matrix
            # always do this for TRs
            if len(subfams_count[subfam]) == 1:
                consensus_pos = consensus_matrix[row_index, col_index]
                consensus_matrix_collapse[
                    subfams_collapse_temp[subfam], col_index
                ] = consensus_matrix[row_index, col_index]
                strand_matrix_collapse[
                    subfams_collapse_temp[subfam], col_index
                ] = strands[row_index]
                support_matrix_collapse[
                    subfams_collapse_temp[subfam], col_index
                ] = support_matrix[row_index, col_index]

                subfam_alignments_collapse[
                    subfam, col_index + start_all - 1
                ] = (row_index, consensus_pos)

            active_cells_collapsed_set[col_index].add(
                subfams_collapse_temp[subfam]
            )

    # convert all sets into lists for easier use later
    active_cells_collapse: Dict[int, List[int]] = {}
    for key in active_cells_collapsed_set:
        active_cells_collapse[key] = list(active_cells_collapsed_set[key])

    for subfam in subfams_dp:
        dp_rows = subfams_count[subfam]  # alignment rows for this subfam
        dp_active_cols: Dict[int, List[int]] = {}

        for dp_row_index in dp_rows:
            dp_col_start = starts[dp_row_index] - start_all + 1
            dp_col_stop = stops[dp_row_index] - start_all + 2
            for dp_col_index in range(dp_col_start, dp_col_stop):
                dp_active_cols.setdefault(dp_col_index, []).append(dp_row_index)

        dp_non_empty_cols = list(dp_active_cols.keys())
        dp_non_empty_cols.sort()

        dp_region_starts = []
        dp_region_stops = []
        prev_start = 0
        count = 0
        dp_region_rows = set()
        dp_active_rows: List[List[int]] = []

        # start at index 1 to look back at index 0
        for dp_col_index in range(1, len(dp_non_empty_cols)):
            dp_region_rows.update(
                dp_active_cols[dp_non_empty_cols[dp_col_index]]
            )

            if (
                dp_non_empty_cols[dp_col_index]
                != dp_non_empty_cols[dp_col_index - 1] + 1
            ):
                # new chrom region for this subfam
                # prev and cur active cols aren't sequential
                dp_region_starts.append(prev_start)
                dp_region_stops.append(dp_col_index - 1)
                prev_start = dp_col_index
                dp_active_rows.append(list(dp_region_rows))
                dp_region_rows = set()
                count += 1
        dp_region_starts.append(prev_start)
        dp_region_stops.append(len(dp_non_empty_cols) - 1)
        dp_active_rows.append(list(dp_region_rows))

        for region_index in range(len(dp_region_starts)):
            region_start = dp_region_starts[region_index]
            region_stop = dp_region_stops[region_index]
            non_empty_region = dp_non_empty_cols[region_start : region_stop + 1]
            if len(dp_active_rows[region_index]) > 1:
                # do dp
                collapse_path = dp_for_collapse(
                    dp_active_rows[region_index],
                    dp_active_cols,
                    support_matrix,
                    non_empty_region,
                )

                for dp_col_index in range(len(non_empty_region)):
                    collapse_col = non_empty_region[dp_col_index]
                    collapse_row = collapse_path[dp_col_index]
                    consensus_pos = consensus_matrix[collapse_row, collapse_col]

                    subfam_alignments_collapse[
                        subfam, collapse_col + start_all - 1
                    ] = (collapse_row, consensus_pos)
                    consensus_matrix_collapse[
                        subfams_collapse_temp[subfam], collapse_col
                    ] = consensus_pos
                    support_matrix_collapse[
                        subfams_collapse_temp[subfam], collapse_col
                    ] = support_matrix[collapse_row, collapse_col]
                    strand_matrix_collapse[
                        subfams_collapse_temp[subfam], collapse_col
                    ] = strands[collapse_row]
            else:
                # dont do dp
                for dp_col_index in range(len(non_empty_region)):
                    collapse_col = non_empty_region[dp_col_index]
                    row_index = dp_active_rows[region_index][0]

                    consensus_pos = consensus_matrix[row_index, collapse_col]
                    # only one row to pick from
                    collapse_row = dp_active_rows[region_index][0]

                    subfam_alignments_collapse[
                        subfam, collapse_col + start_all - 1
                    ] = (collapse_row, consensus_pos)

                    consensus_matrix_collapse[
                        subfams_collapse_temp[subfam], collapse_col
                    ] = consensus_pos
                    strand_matrix_collapse[
                        subfams_collapse_temp[subfam], collapse_col
                    ] = strands[row_index]
                    support_matrix_collapse[
                        subfams_collapse_temp[subfam], collapse_col
                    ] = support_matrix[row_index, collapse_col]
    row_count_update: int = len(subfams_collapse)

    return CollapsedMatrices(
        row_count_update,
        consensus_matrix_collapse,
        strand_matrix_collapse,
        support_matrix_collapse,
        subfams_collapse,
        active_cells_collapse,
        subfams_collapse_temp,
        subfam_alignments_collapse,
    )



class FileFormatException(Exception):
    """
    An exception that should be raised whenever we attempt to parse
    an input file and find that it is in the incorrect format.
    """

    path: str
    line_number: int
    message: str

    def __init__(self, path: str, line_number: int, message: str = ""):
        self.path = path
        self.line_number = line_number
        self.message = message



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['polyA']

package_data = \
{'': ['*']}

entry_points = \
{'console_scripts': ['cm_to_stockholm = polyA.converters:cm_to_stockholm',
                     'polyA = polyA.__main__:main']}

setup(name='polyA',
      version='1.0.0',
      description='PolyA is a sequence annotation adjudicator.',
      author='Kaitlin Carey',
      author_email='kaitlin1.carey@umontana.edu',
      url='https://github.com/TravisWheelerLab/polyA',
      packages=packages,
      package_data=package_data,
      entry_points=entry_points,
     )



from argparse import ArgumentParser, Namespace
from typing import List, Literal, Optional

from . import __version__
from .constants import (
    DEFAULT_CHUNK_SIZE,
    DEFAULT_SHARD_GAP,
)


class Options:
    """
    A typed container to hold program options and parameters.

    >>> o = Options()
    >>> o.log_file_path
    ''
    >>> o = Options(["", "", "--log-file", "foo.txt"])
    >>> o.log_file_path
    'foo.txt'
    """

    alignments_file_path: str
    sub_matrices_path: str

    # -----------------------
    # Algorithm Configuration
    # -----------------------

    chunk_size: int
    confidence: bool
    prior_counts_path: str
    shard_gap: int
    sequence_file_path: str
    ultra_data_path: str
    complexity_adjustment: bool

    # -------------------
    # Helper applications
    # -------------------

    easel_path: str
    ultra_path: str

    # --------------------
    # Output configuration
    # --------------------

    ids: bool
    log_file_path: str
    log_level: Literal["debug", "verbose", "normal", "quiet"]
    matrix_position: bool
    output_path: str
    output_to_file: bool
    sequence_position: bool
    soda: bool

    def __init__(self, args: Optional[List[str]] = None) -> None:
        parser = ArgumentParser(
            description="PolyA sequence adjudication tool",
            prog=__package__,
        )

        parser.add_argument(
            "alignments_file_path",
            metavar="ALIGNMENTS",
            help="alignments file in Stockholm format",
        )
        parser.add_argument(
            "sub_matrices_path",
            metavar="MATRICES",
            help="substitution matrices file in PolyA matrix format",
        )

        parser.add_argument(
            "-v",
            "--version",
            action="version",
            version=__version__,
            help="show version and exit",
        )

        parser.add_argument(
            "--chunk-size",
            type=int,
            default=DEFAULT_CHUNK_SIZE,
            help="size of the window in base pairs analyzed together",
        )
        parser.add_argument(
            "--confidence",
            action="store_true",
            default=False,
            help="run the confidence calculation and then exit",
        )
        parser.add_argument(
            "--prior-counts",
            metavar="FILE",
            default="",
            help="file containing query genomic counts",
        )
        parser.add_argument(
            "--shard-gap",
            type=int,
            default=DEFAULT_SHARD_GAP,
            help="maximum alignment gap before sharding occurs",
        )
        parser.add_argument(
            "--sequences",
            metavar="SEQS",
            default="",
            help="fasta file for running ULTRA",
        )
        parser.add_argument(
            "--ultra-data",
            metavar="FILE",
            default="",
            help="file of the output from ULTRA",
        )

        parser.add_argument(
            "--easel-path",
            metavar="BIN",
            default="",
            help="path to the esl_scorematrix program, if necessary (assumed to be in PATH)",
        )
        parser.add_argument(
            "--ultra-path",
            metavar="BIN",
            default="ultra",
            help="path to the ULTRA binary to use, if necessary (assumed to be in PATH)",
        )

        parser.add_argument(
            "--ids",
            action="store_true",
            default=False,
            help="include internal (random) annotation IDs in output",
        )
        parser.add_argument(
            "--log-file",
            metavar="LOG",
            default="",
            help="file to store log output in, defaults to stderr",
        )
        parser.add_argument(
            "--log-level",
            metavar="LEVEL",
            choices=["debug", "verbose", "normal", "quiet"],
            help="logging level to use, 'debug' is the most noisy",
        )
        parser.add_argument(
            "--matrix-position",
            action="store_true",
            default=False,
            help="produce output in terms of the matrix position",
        )
        parser.add_argument(
            "--output-path",
            metavar="PATH",
            default=".",
            help="directory to write output files to, defaults to working directory",
        )
        parser.add_argument(
            "--output-to-file",
            action="store_true",
            default=False,
            help="write output to a file in the output path",
        )
        parser.add_argument(
            "--sequence-position",
            action="store_true",
            default=False,
            help="produce output in terms of the target sequence position",
        )
        parser.add_argument(
            "--soda",
            action="store_true",
            default=False,
            help="write a SODA visualization file to the output directory",
        )
        parser.add_argument(
            "--complexity-adjustment",
            action="store_true",
            default=False,
            help="use complexity adjusted scoring",
        )

        namespace: Namespace
        if args is None:
            namespace = parser.parse_args(args=["", ""])
        else:
            namespace = parser.parse_args(args=args)

        self.alignments_file_path = namespace.alignments_file_path
        self.sub_matrices_path = namespace.sub_matrices_path

        self.chunk_size = namespace.chunk_size
        self.confidence = namespace.confidence
        self.prior_counts_path = namespace.prior_counts
        self.shard_gap = namespace.shard_gap
        self.sequence_file_path = namespace.sequences
        self.ultra_data_path = namespace.ultra_data
        self.complexity_adjustment = namespace.complexity_adjustment

        self.easel_path = namespace.easel_path
        self.ultra_path = namespace.ultra_path

        self.ids = namespace.ids
        self.log_file_path = namespace.log_file
        self.log_level = namespace.log_level
        self.matrix_position = namespace.matrix_position
        self.output_path = namespace.output_path
        self.output_to_file = namespace.output_to_file
        self.sequence_position = namespace.sequence_position
        self.soda = namespace.soda



import json
from sys import stdout
from typing import Dict, List, Optional, TextIO, Tuple, Any

from .matrices import SupportMatrix, SubfamAlignmentsMatrix


class Printer:
    __output_file: TextIO = stdout
    __print_id: bool
    __soda_conf_file: Optional[TextIO]
    __soda_viz_file: Optional[TextIO]
    __use_matrix_position: bool
    __use_sequence_position: bool

    def __init__(
        self,
        output_file: Optional[TextIO] = None,
        print_id: bool = False,
        soda_conf_file: Optional[TextIO] = None,
        soda_viz_file: Optional[TextIO] = None,
        use_matrix_position: bool = False,
        use_sequence_position: bool = False,
    ) -> None:
        if output_file is not None:
            self.__output_file = output_file

        self.__print_id = print_id
        self.__soda_conf_file = soda_conf_file
        self.__soda_viz_file = soda_viz_file
        self.__use_matrix_position = use_matrix_position
        self.__use_sequence_position = use_sequence_position

    @property
    def use_matrix_position(self) -> bool:
        return self.__use_matrix_position

    @property
    def use_sequence_position(self) -> bool:
        return self.__use_sequence_position

    @property
    def use_soda_output(self) -> bool:
        return (
            self.__soda_viz_file is not None
            and self.__soda_conf_file is not None
        )

    def set_soda_files(
        self, viz_file: Optional[TextIO], conf_file: Optional[TextIO]
    ):
        self.__soda_viz_file = viz_file
        self.__soda_conf_file = conf_file

    def print_results_header(self):
        self.__output_file.write("start\tstop\t")

        if self.__print_id:
            self.__output_file.write("id\t")

        self.__output_file.write("name\n")

    def print_results_simple(
        self,
        start: int,
        stop: int,
        node_id: str,
        node_name: str,
    ) -> None:
        """
        A version of the output printer that takes the exact values to output.
        """
        self.__output_file.write(f"{start}\t{stop}\t")

        if self.__print_id:
            self.__output_file.write(f"{node_id}\t")

        self.__output_file.write(f"{node_name}\n")

    def print_results_matrix(
        self,
        changes: List[str],
        tr_changes: Dict[int, str],
        position_changes: List[int],
        columns: List[int],
        ids: List[str],
    ) -> None:
        """
        Print final results. Start and stop are in terms of matrix position.
        """
        for index, change in enumerate(changes):
            if changes[index] == "skip":
                continue

            position_change = position_changes[index]
            position_change_next = position_changes[index + 1]

            start = columns[position_change]
            stop = columns[position_change_next - 1]

            self.__output_file.write(f"{start}\t")
            self.__output_file.write(f"{stop}\t")

            if self.__print_id:
                self.__output_file.write(f"{ids[start]}\t")

            if change == "Tandem Repeat":
                tr_change = tr_changes[index]
                self.__output_file.write(tr_change)
            else:
                self.__output_file.write(change)

            self.__output_file.write("\n")

    def print_results_sequence(
        self,
        start_all: int,
        changes: List[str],
        tr_changes: Dict[int, str],
        position_changes: List[int],
        columns: List[int],
        ids: List[str],
    ) -> None:
        """
        Print final results. Start and stop are in terms of the target sequence.
        """
        for index, change in enumerate(changes):
            if changes[index] == "skip":
                continue

            position_change = position_changes[index]
            position_change_next = position_changes[index + 1]

            start = columns[position_change] + start_all - 1
            stop = columns[position_change_next - 1] + start_all - 1

            self.__output_file.write(f"{start}\t")
            self.__output_file.write(f"{stop}\t")

            if self.__print_id:
                self.__output_file.write(f"{ids[columns[position_change]]}\t")

            if change == "Tandem Repeat":
                tr_change = tr_changes[index]
                self.__output_file.write(tr_change)
            else:
                self.__output_file.write(change)

            self.__output_file.write("\n")

    def print_results_chrom(
        self,
        start_all: int,
        chrom_start: int,
        changes: List[str],
        tr_changes: Dict[int, str],
        position_changes: List[int],
        columns: List[int],
        ids: List[str],
    ) -> None:
        """
        Print final results. Start and stop are in terms of the chromosome /
        target position.
        """

        for index, change in enumerate(changes):
            if change == "skip":
                continue

            position_change = position_changes[index]
            position_change_next = position_changes[index + 1]

            start = columns[position_change] + start_all + chrom_start - 2
            stop = (
                columns[position_change_next - 1] + start_all + chrom_start - 2
            )

            self.__output_file.write(f"{start}\t")
            self.__output_file.write(f"{stop}\t")

            if self.__print_id:
                self.__output_file.write(f"{ids[columns[position_change]]}\t")

            if change == "Tandem Repeat":
                tr_change = tr_changes[index]
                self.__output_file.write(tr_change)
            else:
                self.__output_file.write(change)

            self.__output_file.write("\n")

    def print_results_soda(
        self,
        start_all: int,
        chrom: str,
        chrom_start: int,
        chrom_end: int,
        subfams: List[str],
        changes_orig: List[str],
        tr_consensus_changes: Dict[int, str],
        changes_position_orig: List[int],
        columns_orig: List[int],
        consensus_lengths: Dict[str, int],
        strand_matrix_collapse: Dict[Tuple[int, int], str],
        consensus_matrix_collapse: Dict[Tuple[int, int], int],
        subfams_collapse_index: Dict[str, int],
        node_confidence_orig: Dict[Tuple[str, int], float],
        ids: List[str],
        subfam_alignments: List[str],
        chrom_alignments: List[str],
        consensus_starts: List[int],
        consensus_stops: List[int],
        alignments_matrix: SubfamAlignmentsMatrix,
        matrix: SupportMatrix,
        subfams_collapse: List[str],
        num_col: int,
    ) -> None:
        """
        Prints the results in the proper format to input into the SODA visualization
        tool.

        The output format is described here:
        https://genome.ucsc.edu/cgi-bin/hgTables?db=hg38&hgta_group=rep&hgta_track=joinedRmsk&hgta_table=rmskJoinedCurrent&hgta_doSchema=describe+table+schema
        """
        if self.__soda_conf_file is None or self.__soda_viz_file is None:
            return

        current_id: int = 0
        length = len(changes_position_orig) - 1

        # wont print out the results of the same thing twice
        used: List[int] = [1] * length

        json_dict_id: Dict[str, Any] = {}

        json_dict: Dict[str, Any] = {
            "chr": chrom,
            "annotations": [],
            "heatmap": [],
        }

        # heatmap
        subfam_ids: Dict[str, int] = {}
        subfam_subids: Dict[str, int] = {}
        subfam_ids[subfams_collapse[0]] = 0
        subfam_subids[subfams_collapse[0]] = 1
        heatmap_dict = {"name": subfams_collapse[0], "id": 0}
        heatmap_vals = []
        confidence = []
        j: int = 0
        cur_subfam_row: int = 0
        align_start: int = chrom_start - 1
        # skip state values
        while j < num_col:
            heatmap_vals.append(round(matrix[0, j], 3))
            j += 1
        confidence.append({"chromStart": align_start, "values": heatmap_vals})
        heatmap_dict["confidence"] = confidence
        heatmap_dict["alignments"] = []
        json_dict["heatmap"].append(heatmap_dict)
        for k in range(1, len(subfams_collapse)):
            subfam_ids[subfams_collapse[k]] = k
            subfam_subids[subfams_collapse[k]] = 1
            heatmap_dict = {"name": subfams_collapse[k], "id": k}
            heatmap_vals = []
            subfam_rows = []
            confidence = []
            alignments = []
            cur_col = 1
            prev_col = 0
            prev_subfam_row = 0
            align_start = chrom_start
            while cur_col < num_col:
                if (k, cur_col) in matrix:
                    # cur subfam row will be unique
                    subfam_row_consensus = alignments_matrix[
                        subfams_collapse[k], cur_col + start_all - 1
                    ]
                    cur_subfam_row = subfam_row_consensus[0]
                    if (
                        cur_col == prev_col
                        and cur_subfam_row == prev_subfam_row
                    ):
                        # continue to add values
                        heatmap_vals.append(round(matrix[k, cur_col], 3))
                        subfam_rows.append(subfam_row_consensus[1])
                        prev_col += 1
                    else:
                        if len(heatmap_vals) > 0:
                            block_sub_alignment: Dict[str, Any] = {}
                            if subfams_collapse[k] != "Tandem Repeat":
                                block_sub_alignment["id"] = prev_subfam_row
                                block_sub_alignment[
                                    "chrSeq"
                                ] = chrom_alignments[prev_subfam_row]
                                block_sub_alignment[
                                    "famSeq"
                                ] = subfam_alignments[prev_subfam_row]
                                block_sub_alignment["relativeStart"] = abs(
                                    consensus_starts[prev_subfam_row]
                                    - subfam_rows[0]
                                )
                                block_sub_alignment["relativeEnd"] = abs(
                                    consensus_stops[prev_subfam_row]
                                    - subfam_rows[-1]
                                )
                                block_sub_alignment[
                                    "alignStart"
                                ] = consensus_starts[prev_subfam_row]
                                block_sub_alignment[
                                    "alignEnd"
                                ] = consensus_stops[prev_subfam_row]
                                alignments.append(block_sub_alignment)
                            confidence.append(
                                {
                                    "chromStart": align_start,
                                    "values": heatmap_vals,
                                }
                            )
                        heatmap_vals = [round(matrix[k, cur_col], 3)]
                        subfam_rows = [subfam_row_consensus[1]]
                        align_start = chrom_start + cur_col + start_all - 2
                        prev_col = cur_col + 1
                        prev_subfam_row = cur_subfam_row
                cur_col += 1
            if len(heatmap_vals) > 0:
                block_sub_alignment = {}
                if subfams_collapse[k] != "Tandem Repeat":
                    block_sub_alignment["id"] = cur_subfam_row
                    block_sub_alignment["chrSeq"] = chrom_alignments[
                        cur_subfam_row
                    ]
                    block_sub_alignment["famSeq"] = subfam_alignments[
                        cur_subfam_row
                    ]
                    block_sub_alignment["relativeStart"] = abs(
                        consensus_starts[cur_subfam_row] - subfam_rows[0]
                    )
                    block_sub_alignment["relativeEnd"] = abs(
                        consensus_stops[cur_subfam_row] - subfam_rows[-1]
                    )
                    block_sub_alignment["alignStart"] = consensus_starts[
                        cur_subfam_row
                    ]
                    block_sub_alignment["alignEnd"] = consensus_stops[
                        cur_subfam_row
                    ]
                    alignments.append(block_sub_alignment)
                confidence.append(
                    {"chromStart": align_start, "values": heatmap_vals}
                )
            heatmap_dict["confidence"] = confidence
            heatmap_dict["alignments"] = alignments
            json_dict["heatmap"].append(heatmap_dict)

        # annotations
        min_align_start: int = chrom_end
        max_align_end: int = 0
        i = 0
        while i < length:
            sub_id: int = 0

            # Note: PyCharm isn't (as of 2021-06) smart enough to figure
            # out the types here, but they are correct and Mypy gets it right.
            json_dict_subid: Dict[str, List[Tuple[str, float]]] = {}

            if changes_orig[i] != "skip" and used[i]:
                orig_subfam: str = changes_orig[i]
                subfam: str = changes_orig[i]
                strand: str = strand_matrix_collapse[
                    subfams_collapse_index[subfam],
                    columns_orig[changes_position_orig[i]],
                ]

                left_flank: int
                right_flank: int

                if orig_subfam == "Tandem Repeat":
                    subfam = tr_consensus_changes[i]
                    left_flank = 0
                    right_flank = 0
                else:
                    if strand == "-":
                        left_flank = (
                            consensus_lengths[subfam]
                            - consensus_matrix_collapse[
                                subfams_collapse_index[subfam],
                                columns_orig[changes_position_orig[i]],
                            ]
                        )
                        right_flank = (
                            consensus_matrix_collapse[
                                subfams_collapse_index[subfam],
                                columns_orig[changes_position_orig[i + 1] - 1],
                            ]
                            - 1
                        )
                    else:
                        left_flank = (
                            consensus_matrix_collapse[
                                subfams_collapse_index[subfam],
                                columns_orig[changes_position_orig[i]],
                            ]
                            - 1
                        )
                        right_flank = (
                            consensus_lengths[subfam]
                            - consensus_matrix_collapse[
                                subfams_collapse_index[subfam],
                                columns_orig[changes_position_orig[i + 1] - 1],
                            ]
                        )

                align_start = (
                    chrom_start
                    + (columns_orig[changes_position_orig[i]] + start_all)
                    - 2
                )
                feature_start: int = align_start - left_flank
                align_stop: int = (
                    chrom_start
                    + (
                        columns_orig[changes_position_orig[i + 1] - 1]
                        + start_all
                    )
                    - 2
                )
                feature_stop: int = align_stop + right_flank

                block_start_matrix: int = columns_orig[changes_position_orig[i]]

                block_count: int = 3
                current_id += 1

                json_dict_subfam_i: Dict[str, float] = {}

                for subfam_i in range(1, len(subfams)):
                    subfamm = subfams[subfam_i]
                    if (
                        subfamm,
                        i,
                    ) in node_confidence_orig and node_confidence_orig[
                        subfamm, i
                    ] > 0.001:
                        json_dict_subfam_i[subfamm] = node_confidence_orig[
                            subfamm, i
                        ]

                json_dict_subid[str(id) + "-" + str(sub_id)] = sorted(
                    json_dict_subfam_i.items(), key=lambda x: x[1], reverse=True
                )
                sub_id += 1

                block_start: List[str] = []
                block_size: List[str] = []

                block_start.append("-1")
                block_start.append(str(left_flank + 1))
                block_start.append("-1")

                block_size.append(str(left_flank))
                block_size.append(
                    str(
                        columns_orig[changes_position_orig[i + 1] - 1]
                        - columns_orig[changes_position_orig[i]]
                        + 1
                    )
                )
                block_size.append(str(right_flank))

                j = i + 1
                while j < length:
                    if (
                        changes_orig[j] != "skip"
                        and orig_subfam != "Tandem Repeat"
                    ):

                        if (
                            ids[columns_orig[changes_position_orig[i]]]
                            == ids[columns_orig[changes_position_orig[j]]]
                        ):
                            if strand == "-":
                                right_flank = (
                                    consensus_matrix_collapse[
                                        subfams_collapse_index[changes_orig[j]],
                                        columns_orig[
                                            changes_position_orig[j + 1] - 1
                                        ],
                                    ]
                                    - 1
                                )
                            else:
                                right_flank = (
                                    consensus_lengths[subfam]
                                    - consensus_matrix_collapse[
                                        subfams_collapse_index[changes_orig[j]],
                                        columns_orig[
                                            changes_position_orig[j + 1] - 1
                                        ],
                                    ]
                                )

                            del block_size[-1]
                            block_size.append("0")

                            align_stop = chrom_start + (
                                columns_orig[changes_position_orig[j + 1] - 1]
                                + start_all
                            )
                            feature_stop = align_stop + right_flank

                            block_start.append(
                                str(
                                    columns_orig[changes_position_orig[j]]
                                    + 1
                                    - block_start_matrix
                                    + left_flank
                                )
                            )
                            block_start.append("-1")

                            block_size.append(
                                str(
                                    columns_orig[
                                        changes_position_orig[j + 1] - 1
                                    ]
                                    - columns_orig[changes_position_orig[j]]
                                    + 1
                                )
                            )

                            block_size.append(str(right_flank))

                            block_count += 2

                            used[j] = 0

                            json_dict_subfam_j: Dict[str, float] = {}

                            for subfam_i in range(1, len(subfams)):
                                subfamm = subfams[subfam_i]
                                if (
                                    (subfamm, j) in node_confidence_orig
                                    and node_confidence_orig[subfamm, j] > 0.001
                                ):
                                    json_dict_subfam_j[
                                        subfamm
                                    ] = node_confidence_orig[subfamm, j]

                            json_dict_subid[
                                str(id) + "-" + str(sub_id)
                            ] = sorted(
                                json_dict_subfam_j.items(),
                                key=lambda x: x[1],
                                reverse=True,
                            )
                            sub_id += 1

                    j += 1
                json_dict_id[str(id)] = json_dict_subid
                json_annotation: Dict[str, Any] = {
                    "bin": "0",
                    "chrom": chrom,
                    "chromStart": str(feature_start),
                    "chromEnd": str(feature_stop),
                    "name": subfam,
                    "score": "0",
                    "strand": strand,
                    "alignStart": str(align_start),
                    "alignEnd": str(align_stop),
                    "reserved": "0",
                    "blockCount": str(block_count),
                    "blockSizes": block_size,
                    "blockStarts": block_start,
                    "id": (
                        str(subfam_ids[orig_subfam])
                        + "-"
                        + str(subfam_subids[orig_subfam])
                    ),
                }
                subfam_subids[orig_subfam] += 1

                json_dict["annotations"].append(json_annotation)

                if align_start < min_align_start:
                    min_align_start = align_start
                if align_stop > max_align_end:
                    max_align_end = align_stop
            used[i] = 0
            i += 1

        json_dict["chrStart"] = min_align_start
        json_dict["chrEnd"] = max_align_end

        # prints  outfile for SODA viz
        self.__soda_viz_file.write(json.dumps(json_dict))

        # prints json file with confidence values for each annotation
        self.__soda_conf_file.write(json.dumps(json_dict_id))



from math import inf, log
from typing import Dict, List, Tuple

from polyA.matrices import CollapsedMatrices
from polyA.performance import timeit


@timeit()
def fill_probability_matrix(
    same_prob_skip: float,
    same_prob: float,
    change_prob: float,
    change_prob_skip: float,
    columns: List[int],
    collapsed_matrices: CollapsedMatrices,
) -> Tuple[List[float], Dict[Tuple[int, int], int], Dict[Tuple[int, int], int]]:
    """
    Calculates the probability score matrix to find most probable path through
    the support matrix. Fills the origin matrix for easier backtrace.

    The basic algorithm is described below.

    1. Look at all i's in j-1
    2. Multiply by confidence in current cell
    3. If it comes from same i, multiply by the higher probability
    4. Else multiply by the lower probability divided by (numseqs-1) so sum of
       all probs == 1
    5. Return max

    .. note::
        All probabilities are in log space.

    Args:
        same_prob_skip: penalty given to staying in the skip state
        same_prob: penalty given to staying in the same row
        change_prob: penalty given for changing rows
        change_prob_skip: penalty given for changing rows in or out of skip state
        columns: list that holds all non empty columns in matrices
            CollapsedMatrices container

    Returns:
        Tuple:
          1. :code:`col_list`: last column of prob matrix, needed to find max to know where
             to start the backtrace.
          2. :code:`origin_matrix`: Hash implementation of sparse 2D DP matrix. This is a
             collapsed matrix. Holds which cell in previous column the probability in
             the DP matrix came from. Used when doing backtrace through the DP
             matrix.
          3. :code:`same_subfam_change_matrix`: parallel to origin_matrix, if 1 - came from
             same subfam, but got a change probability. When doing backtrace, have to
             note this is same subfam name, but different annotation.

    .. todo:: Add a larger test for this function
    """
    active_cells_collapse = collapsed_matrices.active_rows
    support_matrix_collapse = collapsed_matrices.support_matrix
    strand_matrix_collapse = collapsed_matrices.strand_matrix
    consensus_matrix_collapse = collapsed_matrices.consensus_matrix

    origin_matrix: Dict[Tuple[int, int], int] = {}
    same_subfam_change_matrix: Dict[Tuple[int, int], int] = {}
    # speed up by storing previous column (that was just calculated) in a short list for quicker access
    col_list: List[float] = []

    prev_col_list: List[float] = []
    # first col of prob_matrix is 0s
    for _ in active_cells_collapse[columns[0]]:
        prev_col_list.append(0.0)

    consensus_curr: int
    consensus_prev: int
    strand_curr: str
    strand_prev: str

    for columns_index in range(1, len(columns)):
        curr_column: int = columns[columns_index]
        prev_column: int = columns[columns_index - 1]
        col_list.clear()

        for row_index in active_cells_collapse[curr_column]:
            max_value: float = -inf
            max_index: int = 0
            support_log: float = log(
                support_matrix_collapse[row_index, curr_column]
            )
            same_subfam_change: int = 0  # if 1 - comes from the same row, but gets change prob - add to same_subfam_change_matrix and use later in GetPath()

            strand_curr = strand_matrix_collapse[
                row_index, columns[columns_index]
            ]
            consensus_curr = consensus_matrix_collapse[row_index, curr_column]

            # loop through all the rows in the previous column that have a value
            # active_cells_collapse specifies which rows have a value for each column
            temp_index: int = 0
            for prev_row_index in active_cells_collapse[prev_column]:

                score: float = support_log + prev_col_list[temp_index]
                temp_index += 1
                prob: float = change_prob

                if row_index == 0 or prev_row_index == 0:
                    prob = change_prob_skip + same_prob_skip
                    if row_index == prev_row_index:
                        prob = same_prob_skip
                else:
                    if prev_row_index == row_index:  # staying in same row
                        prob = same_prob

                        if (
                            strand_matrix_collapse[prev_row_index, prev_column]
                            != strand_curr
                        ):
                            prob = change_prob

                        elif (
                            strand_matrix_collapse[prev_row_index, prev_column]
                            == "+"
                        ):
                            if (
                                consensus_matrix_collapse[
                                    prev_row_index, prev_column
                                ]
                                > consensus_curr + 50
                            ):
                                prob = change_prob
                                same_subfam_change = 1
                        else:
                            if (
                                consensus_matrix_collapse[
                                    prev_row_index, prev_column
                                ]
                                + 50
                                < consensus_curr
                            ):
                                prob = change_prob
                                same_subfam_change = 1

                score = score + prob

                if score > max_value:
                    max_value = score
                    max_index = prev_row_index

            col_list.append(max_value)
            origin_matrix[row_index, curr_column] = max_index

            if same_subfam_change == 1 and max_index == row_index:
                same_subfam_change_matrix[row_index, curr_column] = 1

        prev_col_list = col_list.copy()

    return col_list, origin_matrix, same_subfam_change_matrix
